extern const unsigned char splash[42503];
extern const unsigned char wxicon[3382];
extern const unsigned char about_menu_16[380];
extern const unsigned char about_menu_24[501];
extern const unsigned char about_menu_32[567];
extern const unsigned char about_menu_48[1512];
extern const unsigned char about_menu_64[1322];
extern const unsigned char arrow_down_16[153];
extern const unsigned char arrow_down_24[208];
extern const unsigned char arrow_down_32[194];
extern const unsigned char arrow_down_48[386];
extern const unsigned char arrow_down_64[418];
extern const unsigned char arrow_down_stop_16[162];
extern const unsigned char arrow_down_stop_24[225];
extern const unsigned char arrow_down_stop_32[189];
extern const unsigned char arrow_down_stop_48[403];
extern const unsigned char arrow_down_stop_64[413];
extern const unsigned char arrow_sort_16[271];
extern const unsigned char arrow_sort_24[395];
extern const unsigned char arrow_sort_32[353];
extern const unsigned char arrow_sort_48[808];
extern const unsigned char arrow_sort_64[939];
extern const unsigned char arrow_up_16[155];
extern const unsigned char arrow_up_24[213];
extern const unsigned char arrow_up_32[210];
extern const unsigned char arrow_up_48[393];
extern const unsigned char arrow_up_64[440];
extern const unsigned char arrow_up_stop_16[166];
extern const unsigned char arrow_up_stop_24[232];
extern const unsigned char arrow_up_stop_32[192];
extern const unsigned char arrow_up_stop_48[402];
extern const unsigned char arrow_up_stop_64[428];
extern const unsigned char assdraw_16[929];
extern const unsigned char assdraw_24[1752];
extern const unsigned char assdraw_32[3322];
extern const unsigned char assdraw_48[9913];
extern const unsigned char assdraw_64[10590];
extern const unsigned char attach_button_16[236];
extern const unsigned char attach_button_24[346];
extern const unsigned char attach_button_32[555];
extern const unsigned char attach_button_48[1810];
extern const unsigned char attach_button_64[1883];
extern const unsigned char automation_toolbutton_16[622];
extern const unsigned char automation_toolbutton_24[935];
extern const unsigned char automation_toolbutton_32[805];
extern const unsigned char automation_toolbutton_48[2894];
extern const unsigned char automation_toolbutton_64[2763];
extern const unsigned char bugtracker_button_16[457];
extern const unsigned char bugtracker_button_24[724];
extern const unsigned char bugtracker_button_32[1214];
extern const unsigned char bugtracker_button_48[3233];
extern const unsigned char bugtracker_button_64[3673];
extern const unsigned char button_audio_commit_16[338];
extern const unsigned char button_audio_commit_24[521];
extern const unsigned char button_audio_commit_32[620];
extern const unsigned char button_audio_commit_48[1908];
extern const unsigned char button_audio_commit_64[1829];
extern const unsigned char button_audio_goto_16[192];
extern const unsigned char button_audio_goto_24[318];
extern const unsigned char button_audio_goto_32[331];
extern const unsigned char button_audio_goto_48[816];
extern const unsigned char button_audio_goto_64[858];
extern const unsigned char button_bold_16[186];
extern const unsigned char button_bold_24[205];
extern const unsigned char button_bold_32[196];
extern const unsigned char button_bold_48[355];
extern const unsigned char button_bold_64[384];
extern const unsigned char button_color_four_16[355];
extern const unsigned char button_color_four_24[531];
extern const unsigned char button_color_four_32[615];
extern const unsigned char button_color_four_48[1724];
extern const unsigned char button_color_four_64[1845];
extern const unsigned char button_color_one_16[271];
extern const unsigned char button_color_one_24[385];
extern const unsigned char button_color_one_32[449];
extern const unsigned char button_color_one_48[1210];
extern const unsigned char button_color_one_64[1276];
extern const unsigned char button_color_three_16[381];
extern const unsigned char button_color_three_24[556];
extern const unsigned char button_color_three_32[636];
extern const unsigned char button_color_three_48[1801];
extern const unsigned char button_color_three_64[1935];
extern const unsigned char button_color_two_16[269];
extern const unsigned char button_color_two_24[382];
extern const unsigned char button_color_two_32[438];
extern const unsigned char button_color_two_48[1238];
extern const unsigned char button_color_two_64[1289];
extern const unsigned char button_fontname_16[219];
extern const unsigned char button_fontname_24[312];
extern const unsigned char button_fontname_32[291];
extern const unsigned char button_fontname_48[665];
extern const unsigned char button_fontname_64[768];
extern const unsigned char button_italics_16[135];
extern const unsigned char button_italics_24[170];
extern const unsigned char button_italics_32[166];
extern const unsigned char button_italics_48[319];
extern const unsigned char button_italics_64[374];
extern const unsigned char button_leadin_16[137];
extern const unsigned char button_leadin_24[238];
extern const unsigned char button_leadin_32[132];
extern const unsigned char button_leadin_48[309];
extern const unsigned char button_leadin_64[263];
extern const unsigned char button_leadout_16[137];
extern const unsigned char button_leadout_24[236];
extern const unsigned char button_leadout_32[154];
extern const unsigned char button_leadout_48[309];
extern const unsigned char button_leadout_64[262];
extern const unsigned char button_next_16[136];
extern const unsigned char button_next_24[179];
extern const unsigned char button_next_32[178];
extern const unsigned char button_next_48[372];
extern const unsigned char button_next_64[359];
extern const unsigned char button_pause_16[111];
extern const unsigned char button_pause_24[125];
extern const unsigned char button_pause_32[99];
extern const unsigned char button_pause_48[159];
extern const unsigned char button_pause_64[126];
extern const unsigned char button_play_16[114];
extern const unsigned char button_play_24[135];
extern const unsigned char button_play_32[146];
extern const unsigned char button_play_48[254];
extern const unsigned char button_play_64[269];
extern const unsigned char button_playfirstfiveh_16[212];
extern const unsigned char button_playfirstfiveh_24[389];
extern const unsigned char button_playfirstfiveh_32[268];
extern const unsigned char button_playfirstfiveh_48[599];
extern const unsigned char button_playfirstfiveh_64[521];
extern const unsigned char button_playfivehafter_16[205];
extern const unsigned char button_playfivehafter_24[247];
extern const unsigned char button_playfivehafter_32[270];
extern const unsigned char button_playfivehafter_48[589];
extern const unsigned char button_playfivehafter_64[520];
extern const unsigned char button_playfivehbefore_16[198];
extern const unsigned char button_playfivehbefore_24[254];
extern const unsigned char button_playfivehbefore_32[269];
extern const unsigned char button_playfivehbefore_48[595];
extern const unsigned char button_playfivehbefore_64[521];
extern const unsigned char button_playlastfiveh_16[212];
extern const unsigned char button_playlastfiveh_24[388];
extern const unsigned char button_playlastfiveh_32[268];
extern const unsigned char button_playlastfiveh_48[583];
extern const unsigned char button_playlastfiveh_64[523];
extern const unsigned char button_playline_16[175];
extern const unsigned char button_playline_24[308];
extern const unsigned char button_playline_32[252];
extern const unsigned char button_playline_48[538];
extern const unsigned char button_playline_64[505];
extern const unsigned char button_playsel_16[168];
extern const unsigned char button_playsel_24[305];
extern const unsigned char button_playsel_32[254];
extern const unsigned char button_playsel_48[538];
extern const unsigned char button_playsel_64[504];
extern const unsigned char button_playtoend_16[223];
extern const unsigned char button_playtoend_24[302];
extern const unsigned char button_playtoend_32[248];
extern const unsigned char button_playtoend_48[610];
extern const unsigned char button_playtoend_64[515];
extern const unsigned char button_prev_16[140];
extern const unsigned char button_prev_24[179];
extern const unsigned char button_prev_32[196];
extern const unsigned char button_prev_48[389];
extern const unsigned char button_prev_64[394];
extern const unsigned char button_stop_16[95];
extern const unsigned char button_stop_24[108];
extern const unsigned char button_stop_32[94];
extern const unsigned char button_stop_48[129];
extern const unsigned char button_stop_64[121];
extern const unsigned char button_strikeout_16[165];
extern const unsigned char button_strikeout_24[225];
extern const unsigned char button_strikeout_32[216];
extern const unsigned char button_strikeout_48[452];
extern const unsigned char button_strikeout_64[841];
extern const unsigned char button_underline_16[153];
extern const unsigned char button_underline_24[189];
extern const unsigned char button_underline_32[223];
extern const unsigned char button_underline_48[368];
extern const unsigned char button_underline_64[392];
extern const unsigned char close_audio_menu_16[458];
extern const unsigned char close_audio_menu_24[760];
extern const unsigned char close_audio_menu_32[1006];
extern const unsigned char close_audio_menu_48[2615];
extern const unsigned char close_audio_menu_64[2274];
extern const unsigned char close_keyframes_menu_16[534];
extern const unsigned char close_keyframes_menu_24[852];
extern const unsigned char close_keyframes_menu_32[948];
extern const unsigned char close_keyframes_menu_48[2862];
extern const unsigned char close_keyframes_menu_64[1920];
extern const unsigned char close_timecodes_menu_16[607];
extern const unsigned char close_timecodes_menu_24[1020];
extern const unsigned char close_timecodes_menu_32[904];
extern const unsigned char close_timecodes_menu_48[3075];
extern const unsigned char close_timecodes_menu_64[2218];
extern const unsigned char close_video_menu_16[620];
extern const unsigned char close_video_menu_24[937];
extern const unsigned char close_video_menu_32[1175];
extern const unsigned char close_video_menu_48[3938];
extern const unsigned char close_video_menu_64[2748];
extern const unsigned char contents_button_16[665];
extern const unsigned char contents_button_24[1095];
extern const unsigned char contents_button_32[1389];
extern const unsigned char contents_button_48[4332];
extern const unsigned char contents_button_64[3752];
extern const unsigned char copy_button_16[374];
extern const unsigned char copy_button_24[599];
extern const unsigned char copy_button_32[328];
extern const unsigned char copy_button_48[1251];
extern const unsigned char copy_button_64[594];
extern const unsigned char cut_button_16[510];
extern const unsigned char cut_button_24[944];
extern const unsigned char cut_button_32[1264];
extern const unsigned char cut_button_48[3421];
extern const unsigned char cut_button_64[3729];
extern const unsigned char delete_button_16[153];
extern const unsigned char delete_button_24[174];
extern const unsigned char delete_button_32[214];
extern const unsigned char delete_button_48[522];
extern const unsigned char delete_button_64[355];
extern const unsigned char detach_video_menu_16[585];
extern const unsigned char detach_video_menu_24[814];
extern const unsigned char detach_video_menu_32[937];
extern const unsigned char detach_video_menu_48[4229];
extern const unsigned char detach_video_menu_64[2265];
extern const unsigned char export_menu_16[682];
extern const unsigned char export_menu_24[1008];
extern const unsigned char export_menu_32[779];
extern const unsigned char export_menu_48[2503];
extern const unsigned char export_menu_64[2213];
extern const unsigned char eyedropper_tool_16[236];
extern const unsigned char eyedropper_tool_24[369];
extern const unsigned char eyedropper_tool_32[630];
extern const unsigned char eyedropper_tool_48[1696];
extern const unsigned char eyedropper_tool_64[2036];
extern const unsigned char find_button_16[369];
extern const unsigned char find_button_24[527];
extern const unsigned char find_button_32[663];
extern const unsigned char find_button_48[2054];
extern const unsigned char find_button_64[1754];
extern const unsigned char find_next_menu_16[442];
extern const unsigned char find_next_menu_24[600];
extern const unsigned char find_next_menu_32[633];
extern const unsigned char find_next_menu_48[1912];
extern const unsigned char find_next_menu_64[1755];
extern const unsigned char find_replace_menu_16[585];
extern const unsigned char find_replace_menu_24[916];
extern const unsigned char find_replace_menu_32[843];
extern const unsigned char find_replace_menu_48[2636];
extern const unsigned char find_replace_menu_64[2488];
extern const unsigned char font_collector_button_16[452];
extern const unsigned char font_collector_button_24[635];
extern const unsigned char font_collector_button_32[499];
extern const unsigned char font_collector_button_48[1469];
extern const unsigned char font_collector_button_64[1260];
extern const unsigned char forums_button_16[743];
extern const unsigned char forums_button_24[1235];
extern const unsigned char forums_button_32[1767];
extern const unsigned char forums_button_48[5740];
extern const unsigned char forums_button_64[5267];
extern const unsigned char irc_button_16[563];
extern const unsigned char irc_button_24[894];
extern const unsigned char irc_button_32[1166];
extern const unsigned char irc_button_48[3661];
extern const unsigned char irc_button_64[3593];
extern const unsigned char jumpto_button_16[192];
extern const unsigned char jumpto_button_24[318];
extern const unsigned char jumpto_button_32[331];
extern const unsigned char jumpto_button_48[816];
extern const unsigned char jumpto_button_64[858];
extern const unsigned char kara_mode_16[434];
extern const unsigned char kara_mode_24[659];
extern const unsigned char kara_mode_32[867];
extern const unsigned char kara_mode_48[2668];
extern const unsigned char kara_mode_64[2731];
extern const unsigned char kara_split_16[619];
extern const unsigned char kara_split_24[995];
extern const unsigned char kara_split_32[1269];
extern const unsigned char kara_split_48[3514];
extern const unsigned char kara_split_64[3316];
extern const unsigned char kara_split_accept_16[769];
extern const unsigned char kara_split_accept_24[1510];
extern const unsigned char kara_split_accept_32[1736];
extern const unsigned char kara_split_accept_48[5033];
extern const unsigned char kara_split_accept_64[4658];
extern const unsigned char kara_split_cancel_16[713];
extern const unsigned char kara_split_cancel_24[1139];
extern const unsigned char kara_split_cancel_32[1591];
extern const unsigned char kara_split_cancel_48[4016];
extern const unsigned char kara_split_cancel_64[3508];
extern const unsigned char kara_timing_copier_16[639];
extern const unsigned char kara_timing_copier_24[900];
extern const unsigned char kara_timing_copier_32[992];
extern const unsigned char kara_timing_copier_48[3042];
extern const unsigned char kara_timing_copier_64[2807];
extern const unsigned char languages_menu_16[372];
extern const unsigned char languages_menu_24[582];
extern const unsigned char languages_menu_32[725];
extern const unsigned char languages_menu_48[2003];
extern const unsigned char languages_menu_64[2068];
extern const unsigned char new_toolbutton_16[151];
extern const unsigned char new_toolbutton_24[170];
extern const unsigned char new_toolbutton_32[132];
extern const unsigned char new_toolbutton_48[322];
extern const unsigned char new_toolbutton_64[207];
extern const unsigned char new_window_menu_16[297];
extern const unsigned char new_window_menu_24[410];
extern const unsigned char new_window_menu_32[393];
extern const unsigned char new_window_menu_48[1152];
extern const unsigned char new_window_menu_64[820];
extern const unsigned char open_audio_from_video_menu_16[626];
extern const unsigned char open_audio_from_video_menu_24[1070];
extern const unsigned char open_audio_from_video_menu_32[1241];
extern const unsigned char open_audio_from_video_menu_48[6684];
extern const unsigned char open_audio_from_video_menu_64[4867];
extern const unsigned char open_audio_menu_16[495];
extern const unsigned char open_audio_menu_24[716];
extern const unsigned char open_audio_menu_32[580];
extern const unsigned char open_audio_menu_48[1885];
extern const unsigned char open_audio_menu_64[1136];
extern const unsigned char open_keyframes_menu_16[410];
extern const unsigned char open_keyframes_menu_24[547];
extern const unsigned char open_keyframes_menu_32[503];
extern const unsigned char open_keyframes_menu_48[1932];
extern const unsigned char open_keyframes_menu_64[1026];
extern const unsigned char open_timecodes_menu_16[496];
extern const unsigned char open_timecodes_menu_24[753];
extern const unsigned char open_timecodes_menu_32[567];
extern const unsigned char open_timecodes_menu_48[2356];
extern const unsigned char open_timecodes_menu_64[1507];
extern const unsigned char open_toolbutton_16[333];
extern const unsigned char open_toolbutton_24[522];
extern const unsigned char open_toolbutton_32[481];
extern const unsigned char open_toolbutton_48[1483];
extern const unsigned char open_toolbutton_64[1065];
extern const unsigned char open_video_menu_16[493];
extern const unsigned char open_video_menu_24[644];
extern const unsigned char open_video_menu_32[738];
extern const unsigned char open_video_menu_48[2850];
extern const unsigned char open_video_menu_64[1729];
extern const unsigned char open_with_toolbutton_16[470];
extern const unsigned char open_with_toolbutton_24[604];
extern const unsigned char open_with_toolbutton_32[562];
extern const unsigned char open_with_toolbutton_48[1767];
extern const unsigned char open_with_toolbutton_64[1209];
extern const unsigned char options_button_16[574];
extern const unsigned char options_button_24[821];
extern const unsigned char options_button_32[681];
extern const unsigned char options_button_48[1818];
extern const unsigned char options_button_64[1594];
extern const unsigned char paste_button_16[450];
extern const unsigned char paste_button_24[561];
extern const unsigned char paste_button_32[567];
extern const unsigned char paste_button_48[2070];
extern const unsigned char paste_button_64[1621];
extern const unsigned char properties_toolbutton_16[284];
extern const unsigned char properties_toolbutton_24[502];
extern const unsigned char properties_toolbutton_32[342];
extern const unsigned char properties_toolbutton_48[908];
extern const unsigned char properties_toolbutton_64[748];
extern const unsigned char redo_button_16[269];
extern const unsigned char redo_button_24[353];
extern const unsigned char redo_button_32[489];
extern const unsigned char redo_button_48[1365];
extern const unsigned char redo_button_64[1267];
extern const unsigned char resample_toolbutton_16[207];
extern const unsigned char resample_toolbutton_24[247];
extern const unsigned char resample_toolbutton_32[178];
extern const unsigned char resample_toolbutton_48[406];
extern const unsigned char resample_toolbutton_64[246];
extern const unsigned char save_as_toolbutton_16[400];
extern const unsigned char save_as_toolbutton_24[493];
extern const unsigned char save_as_toolbutton_32[284];
extern const unsigned char save_as_toolbutton_48[921];
extern const unsigned char save_as_toolbutton_64[426];
extern const unsigned char save_keyframes_menu_16[572];
extern const unsigned char save_keyframes_menu_24[758];
extern const unsigned char save_keyframes_menu_32[568];
extern const unsigned char save_keyframes_menu_48[2269];
extern const unsigned char save_keyframes_menu_64[1190];
extern const unsigned char save_timecodes_menu_16[601];
extern const unsigned char save_timecodes_menu_24[882];
extern const unsigned char save_timecodes_menu_32[586];
extern const unsigned char save_timecodes_menu_48[2344];
extern const unsigned char save_timecodes_menu_64[1461];
extern const unsigned char save_toolbutton_16[333];
extern const unsigned char save_toolbutton_24[415];
extern const unsigned char save_toolbutton_32[207];
extern const unsigned char save_toolbutton_48[612];
extern const unsigned char save_toolbutton_64[304];
extern const unsigned char select_lines_button_16[297];
extern const unsigned char select_lines_button_24[444];
extern const unsigned char select_lines_button_32[260];
extern const unsigned char select_lines_button_48[795];
extern const unsigned char select_lines_button_64[442];
extern const unsigned char select_visible_button_16[404];
extern const unsigned char select_visible_button_24[536];
extern const unsigned char select_visible_button_32[303];
extern const unsigned char select_visible_button_48[962];
extern const unsigned char select_visible_button_64[485];
extern const unsigned char shift_times_toolbutton_16[395];
extern const unsigned char shift_times_toolbutton_24[435];
extern const unsigned char shift_times_toolbutton_32[438];
extern const unsigned char shift_times_toolbutton_48[1118];
extern const unsigned char shift_times_toolbutton_64[1006];
extern const unsigned char shift_to_frame_16[296];
extern const unsigned char shift_to_frame_24[404];
extern const unsigned char shift_to_frame_32[320];
extern const unsigned char shift_to_frame_48[912];
extern const unsigned char shift_to_frame_64[810];
extern const unsigned char show_video_details_menu_16[579];
extern const unsigned char show_video_details_menu_24[988];
extern const unsigned char show_video_details_menu_32[1239];
extern const unsigned char show_video_details_menu_48[5396];
extern const unsigned char show_video_details_menu_64[3559];
extern const unsigned char snap_subs_to_scene_16[369];
extern const unsigned char snap_subs_to_scene_24[432];
extern const unsigned char snap_subs_to_scene_32[432];
extern const unsigned char snap_subs_to_scene_48[1147];
extern const unsigned char snap_subs_to_scene_64[733];
extern const unsigned char spellcheck_toolbutton_16[554];
extern const unsigned char spellcheck_toolbutton_24[858];
extern const unsigned char spellcheck_toolbutton_32[1136];
extern const unsigned char spellcheck_toolbutton_48[3503];
extern const unsigned char spellcheck_toolbutton_64[3297];
extern const unsigned char style_toolbutton_16[506];
extern const unsigned char style_toolbutton_24[811];
extern const unsigned char style_toolbutton_32[854];
extern const unsigned char style_toolbutton_48[2658];
extern const unsigned char style_toolbutton_64[2584];
extern const unsigned char styling_toolbutton_16[349];
extern const unsigned char styling_toolbutton_24[550];
extern const unsigned char styling_toolbutton_32[811];
extern const unsigned char styling_toolbutton_48[2416];
extern const unsigned char styling_toolbutton_64[2504];
extern const unsigned char subend_to_video_16[496];
extern const unsigned char subend_to_video_24[701];
extern const unsigned char subend_to_video_32[826];
extern const unsigned char subend_to_video_48[3749];
extern const unsigned char subend_to_video_64[2204];
extern const unsigned char substart_to_video_16[499];
extern const unsigned char substart_to_video_24[695];
extern const unsigned char substart_to_video_32[844];
extern const unsigned char substart_to_video_48[3845];
extern const unsigned char substart_to_video_64[2199];
extern const unsigned char timing_processor_toolbutton_16[665];
extern const unsigned char timing_processor_toolbutton_24[1032];
extern const unsigned char timing_processor_toolbutton_32[1333];
extern const unsigned char timing_processor_toolbutton_48[3816];
extern const unsigned char timing_processor_toolbutton_64[3439];
extern const unsigned char toggle_audio_autocommit_16[435];
extern const unsigned char toggle_audio_autocommit_24[658];
extern const unsigned char toggle_audio_autocommit_32[952];
extern const unsigned char toggle_audio_autocommit_48[2891];
extern const unsigned char toggle_audio_autocommit_64[2887];
extern const unsigned char toggle_audio_autoscroll_16[339];
extern const unsigned char toggle_audio_autoscroll_24[448];
extern const unsigned char toggle_audio_autoscroll_32[705];
extern const unsigned char toggle_audio_autoscroll_48[1992];
extern const unsigned char toggle_audio_autoscroll_64[1994];
extern const unsigned char toggle_audio_link_16[166];
extern const unsigned char toggle_audio_link_24[211];
extern const unsigned char toggle_audio_link_32[137];
extern const unsigned char toggle_audio_link_48[283];
extern const unsigned char toggle_audio_link_64[175];
extern const unsigned char toggle_audio_medusa_16[362];
extern const unsigned char toggle_audio_medusa_24[469];
extern const unsigned char toggle_audio_medusa_32[282];
extern const unsigned char toggle_audio_medusa_48[944];
extern const unsigned char toggle_audio_medusa_64[455];
extern const unsigned char toggle_audio_nextcommit_16[462];
extern const unsigned char toggle_audio_nextcommit_24[748];
extern const unsigned char toggle_audio_nextcommit_32[712];
extern const unsigned char toggle_audio_nextcommit_48[2767];
extern const unsigned char toggle_audio_nextcommit_64[2697];
extern const unsigned char toggle_audio_spectrum_16[628];
extern const unsigned char toggle_audio_spectrum_24[980];
extern const unsigned char toggle_audio_spectrum_32[1693];
extern const unsigned char toggle_audio_spectrum_48[5343];
extern const unsigned char toggle_audio_spectrum_64[4634];
extern const unsigned char toggle_tag_hiding_16[295];
extern const unsigned char toggle_tag_hiding_24[447];
extern const unsigned char toggle_tag_hiding_32[650];
extern const unsigned char toggle_tag_hiding_48[1850];
extern const unsigned char toggle_tag_hiding_64[1958];
extern const unsigned char toggle_video_autoscroll_16[589];
extern const unsigned char toggle_video_autoscroll_24[890];
extern const unsigned char toggle_video_autoscroll_32[1424];
extern const unsigned char toggle_video_autoscroll_48[5015];
extern const unsigned char toggle_video_autoscroll_64[3801];
extern const unsigned char translation_toolbutton_16[497];
extern const unsigned char translation_toolbutton_24[837];
extern const unsigned char translation_toolbutton_32[1279];
extern const unsigned char translation_toolbutton_48[3237];
extern const unsigned char translation_toolbutton_64[3439];
extern const unsigned char undo_button_16[264];
extern const unsigned char undo_button_24[362];
extern const unsigned char undo_button_32[455];
extern const unsigned char undo_button_48[1372];
extern const unsigned char undo_button_64[1324];
extern const unsigned char use_dummy_video_menu_16[245];
extern const unsigned char use_dummy_video_menu_24[464];
extern const unsigned char use_dummy_video_menu_32[260];
extern const unsigned char use_dummy_video_menu_48[907];
extern const unsigned char use_dummy_video_menu_64[645];
extern const unsigned char video_to_subend_16[544];
extern const unsigned char video_to_subend_24[824];
extern const unsigned char video_to_subend_32[1015];
extern const unsigned char video_to_subend_48[4200];
extern const unsigned char video_to_subend_64[2633];
extern const unsigned char video_to_substart_16[555];
extern const unsigned char video_to_substart_24[836];
extern const unsigned char video_to_substart_32[1045];
extern const unsigned char video_to_substart_48[4299];
extern const unsigned char video_to_substart_64[2656];
extern const unsigned char visual_clip_16[379];
extern const unsigned char visual_clip_24[393];
extern const unsigned char visual_clip_32[444];
extern const unsigned char visual_clip_48[1210];
extern const unsigned char visual_clip_64[1058];
extern const unsigned char visual_help_16[448];
extern const unsigned char visual_help_24[543];
extern const unsigned char visual_help_32[833];
extern const unsigned char visual_help_48[2281];
extern const unsigned char visual_help_64[2526];
extern const unsigned char visual_move_16[300];
extern const unsigned char visual_move_24[406];
extern const unsigned char visual_move_32[447];
extern const unsigned char visual_move_48[1029];
extern const unsigned char visual_move_64[767];
extern const unsigned char visual_move_conv_move_16[228];
extern const unsigned char visual_move_conv_move_24[353];
extern const unsigned char visual_move_conv_move_32[460];
extern const unsigned char visual_move_conv_move_48[1151];
extern const unsigned char visual_move_conv_move_64[1069];
extern const unsigned char visual_move_conv_pos_16[354];
extern const unsigned char visual_move_conv_pos_24[439];
extern const unsigned char visual_move_conv_pos_32[337];
extern const unsigned char visual_move_conv_pos_48[872];
extern const unsigned char visual_move_conv_pos_64[725];
extern const unsigned char visual_rotatexy_16[559];
extern const unsigned char visual_rotatexy_24[941];
extern const unsigned char visual_rotatexy_32[1345];
extern const unsigned char visual_rotatexy_48[3551];
extern const unsigned char visual_rotatexy_64[3785];
extern const unsigned char visual_rotatez_16[414];
extern const unsigned char visual_rotatez_24[675];
extern const unsigned char visual_rotatez_32[907];
extern const unsigned char visual_rotatez_48[2217];
extern const unsigned char visual_rotatez_64[2185];
extern const unsigned char visual_scale_16[207];
extern const unsigned char visual_scale_24[247];
extern const unsigned char visual_scale_32[229];
extern const unsigned char visual_scale_48[436];
extern const unsigned char visual_scale_64[387];
extern const unsigned char visual_standard_16[258];
extern const unsigned char visual_standard_24[326];
extern const unsigned char visual_standard_32[425];
extern const unsigned char visual_standard_48[1055];
extern const unsigned char visual_standard_64[961];
extern const unsigned char visual_vector_clip_16[585];
extern const unsigned char visual_vector_clip_24[945];
extern const unsigned char visual_vector_clip_32[1116];
extern const unsigned char visual_vector_clip_48[3738];
extern const unsigned char visual_vector_clip_64[3345];
extern const unsigned char visual_vector_clip_bicubic_16[307];
extern const unsigned char visual_vector_clip_bicubic_24[576];
extern const unsigned char visual_vector_clip_bicubic_32[645];
extern const unsigned char visual_vector_clip_bicubic_48[1755];
extern const unsigned char visual_vector_clip_bicubic_64[1732];
extern const unsigned char visual_vector_clip_convert_16[398];
extern const unsigned char visual_vector_clip_convert_24[595];
extern const unsigned char visual_vector_clip_convert_32[682];
extern const unsigned char visual_vector_clip_convert_48[1690];
extern const unsigned char visual_vector_clip_convert_64[1491];
extern const unsigned char visual_vector_clip_drag_16[369];
extern const unsigned char visual_vector_clip_drag_24[544];
extern const unsigned char visual_vector_clip_drag_32[575];
extern const unsigned char visual_vector_clip_drag_48[1418];
extern const unsigned char visual_vector_clip_drag_64[1115];
extern const unsigned char visual_vector_clip_freehand_16[306];
extern const unsigned char visual_vector_clip_freehand_24[400];
extern const unsigned char visual_vector_clip_freehand_32[525];
extern const unsigned char visual_vector_clip_freehand_48[1692];
extern const unsigned char visual_vector_clip_freehand_64[1419];
extern const unsigned char visual_vector_clip_freehand_smooth_16[394];
extern const unsigned char visual_vector_clip_freehand_smooth_24[579];
extern const unsigned char visual_vector_clip_freehand_smooth_32[745];
extern const unsigned char visual_vector_clip_freehand_smooth_48[2304];
extern const unsigned char visual_vector_clip_freehand_smooth_64[2270];
extern const unsigned char visual_vector_clip_insert_16[343];
extern const unsigned char visual_vector_clip_insert_24[515];
extern const unsigned char visual_vector_clip_insert_32[596];
extern const unsigned char visual_vector_clip_insert_48[1803];
extern const unsigned char visual_vector_clip_insert_64[1634];
extern const unsigned char visual_vector_clip_line_16[147];
extern const unsigned char visual_vector_clip_line_24[238];
extern const unsigned char visual_vector_clip_line_32[190];
extern const unsigned char visual_vector_clip_line_48[414];
extern const unsigned char visual_vector_clip_line_64[405];
extern const unsigned char visual_vector_clip_remove_16[429];
extern const unsigned char visual_vector_clip_remove_24[761];
extern const unsigned char visual_vector_clip_remove_32[773];
extern const unsigned char visual_vector_clip_remove_48[2164];
extern const unsigned char visual_vector_clip_remove_64[1714];
extern const unsigned char website_button_16[733];
extern const unsigned char website_button_24[1219];
extern const unsigned char website_button_32[2478];
extern const unsigned char website_button_48[7307];
extern const unsigned char website_button_64[7293];
extern const unsigned char zoom_in_button_16[462];
extern const unsigned char zoom_in_button_24[883];
extern const unsigned char zoom_in_button_32[1424];
extern const unsigned char zoom_in_button_48[3204];
extern const unsigned char zoom_in_button_64[3864];
extern const unsigned char zoom_out_button_16[466];
extern const unsigned char zoom_out_button_24[886];
extern const unsigned char zoom_out_button_32[1462];
extern const unsigned char zoom_out_button_48[3163];
extern const unsigned char zoom_out_button_64[3789];
