extern const unsigned char default_config[12536];
extern const unsigned char default_config_platform[47];
extern const unsigned char default_hotkey[5207];
extern const unsigned char default_menu[8774];
extern const unsigned char default_mru[120];
extern const unsigned char default_toolbar[1924];
